/* written by John Doug Reynolds, October 2001 */

#include "rgsmcrgen.h"
#include "calib.h"
#include <rgsPixelsLib.h>
#include <CalServer.h>
#include <cstdio>

using namespace rgsPixelsLib;

#define SET(name,value,label) *array->addAttribute(name,label) = value

static void copy( const Spectrum &s, unsigned length, float *&ptr, float &max )
{
   float sum = 0;
   unsigned i = min( (unsigned) s.size(), length );
   Spectrum::const_iterator iter = s.begin();
   while ( i-- ) sum += (*ptr++ = *iter++);
   if ( length > s.size() ) for ( i = length - s.size(); i--; *ptr++ = 0 );
   if ( sum > max ) max = sum;
}

static void standardize( struct ArrayData *data, float normalizer )
{
   float *ptr = data->real32Data();
   for ( unsigned k = data->size()[0]; k--; )
      for ( unsigned j = data->size()[1]; j--; ) {
	 float sum = 0;
	 for ( unsigned i = data->size()[2]; i--; ++ptr )
	    *ptr = normalizer * (sum += *ptr);
      }
}

static float veclen( float x, float y, float z )
{
   return sqrt( x*x + y*y + z*z );
}

void rgsmcrgen()
{
   string dataname = stringParameter("set");
   string filename = stringParameter("file");

   struct GridSpan  phiGrid( realParameter("phimin"),
			     realParameter("phimax"),
			     intParameter("phibins")
			     );
   struct GridSpan  psiGrid( realParameter("psimin"),
			     realParameter("psimax"),
			     intParameter("psibins")
			     );
   struct GridSpan wav1Grid( realParameter("wavmin"),
			     realParameter("wavmax"),
			     intParameter("wavbetabins")
			     );
   struct GridSpan wav2Grid( realParameter("wavmin"),
			     realParameter("wavmax"),
			     intParameter("wavphbins")
			     );

   string::size_type suffix = dataname.rfind('.');
   if ( suffix == string::npos ) suffix = dataname.size();

   if ( filename == "" )
      if ( suffix >= 10 )
	 (filename=dataname).replace( suffix-10, 6, "tensor" );
      else
	 filename = "mcr.fit";

   // initializations

   struct DataSet *dataset = dataSetServer->open( dataname, Dal::Read );

   string instrument = dataset->attribute("INSTRUME")->asString();
   calServer.setInstrument( Xmm::toInstrument(instrument) );

   struct MiscDataServer *misc = 0; misc = calServer.getAtom(misc);
   const string lincoord("LINCOORD:LINCOORD%");
   float fOverL =
      misc->get("FOCAL_LENGTH") /
      veclen( misc->get(lincoord+"G_TX") - misc->get(lincoord+"CC2_TX"),
	      misc->get(lincoord+"G_TY") - misc->get(lincoord+"CC2_TY"),
	      misc->get(lincoord+"G_TZ") - misc->get(lincoord+"CC2_TZ") );
   calServer.releaseAtom(misc);

   const Array *emap = dataset->array("PRIMARY");
   struct GridSpan betaGrid = gridspan(emap,1);
   struct GridSpan xdspGrid = gridspan(emap,2);

   longVec size(3);
   struct Array *array;
   struct ArrayData *data;
   float *rowPtr, scale;

   struct DataSet *mcrset = dataSetServer->open( filename, Dal::Create );

   mcrset->addAttribute( dataset->attribute("TELESCOP") );
   mcrset->addAttribute( dataset->attribute("INSTRUME") );

   //AI: Due to the change in the Quantum eff CCF file, now
   //we have to pass the order to the calibInit function to apply the
   //correct CCD correction 
   //AI: Wavelength scale. Start
   //Due to the wavelength modifications, we also need the betaGrid twice to
   //simulate the betaChannelGrid.
   calibInit( betaGrid, betaGrid,1);
   //AI: Wavelength scale. End
   dataSetServer->close( dataset );

   // BETA_RESPONSE tensors (small and large angle scattering)

   for ( unsigned large = 0; large < 2; ++large ) {
      string extname = large ? "BETA_RESPONSE_LA" : "BETA_RESPONSE_SA";

      size[0] = phiGrid.length();
      size[1] = wav1Grid.length();
      size[2] = betaGrid.length();
      array = mcrset->addArray( extname, Dal::Real32,size,"",
				"response(beta,lambda,phi)" );

      SET( "BETA_REF", betaGrid[betaGrid.min], "midpoint of bin at origin" );
      SET( "BETA_WID", betaGrid.stride, "uniform bin width" );
      SET( "WAV_REF", wav1Grid[wav1Grid.min], "midpoint of bin at origin" );
      SET( "WAV_WID", wav1Grid.stride, "uniform bin width" );
      SET( "PHI_REF", phiGrid[phiGrid.min], "midpoint of bin at origin" );
      SET( "PHI_WID", phiGrid.stride, "uniform bin width" );

      data = array->data();
      rowPtr = data->real32Data();
      scale = 0;

      vector<float> lsf( betaGrid.length() ), component( betaGrid.length() );
      for ( int phi = phiGrid.min; phi <= phiGrid.max; ++phi ) {
	 float delta_phi = phiGrid[phi];
	 float alpha = calAlpha0() + fOverL * delta_phi;

	 if ( delta_phi > 0 )
	    resetSourcePosition( (180/M_PI) * 3600 * delta_phi, M_PI );
	 else
	    resetSourcePosition( -(180/M_PI) * 3600 * delta_phi, 0 );

	 for ( int wav = wav1Grid.min; wav <= wav1Grid.max; ++wav ) {
	    size_t i, bgSize = betaGrid.length();
	    float energy = hc/wav1Grid[wav];

	    for ( i = 0; i < bgSize; ++i ) lsf[i] = 0;

	    for ( int order = 0; order < 6; ++order ) {
	       if ( large )
		  prfBroad( component,energy, energy, alpha, -order );
	       else
		  prfNarrow( component, energy, energy, alpha, -order, 3 );

	       for ( i = 0; i < bgSize; ++i ) lsf[i] += component[i];
	    }

	    struct Spectrum dist( bgSize );
	    for ( i = 0; i < bgSize; ++i ) dist[i] = lsf[i] * vignetting[i];

	    copy( dist, size[2], rowPtr, scale );
	 }
      }

      SET( "EFF-AREA", scale, "inverse normalizer (cm^2)" );
      standardize( data, 1/scale );
      delete data;
   }

   // XDSP_RESPONSE_SA tensor (small angle scattering)

   size[0] = psiGrid.length();
   size[1] = betaGrid.length();
   size[2] = xdspGrid.length();
   array = mcrset->addArray( "XDSP_RESPONSE_SA", Dal::Real32, size, "",
			     "response(xdsp,beta,psi)" );

   SET( "XDSP_REF", xdspGrid[xdspGrid.min], "midpoint of bin at origin" );
   SET( "XDSP_WID", xdspGrid.stride, "uniform bin width" );
   SET( "BETA_REF", betaGrid[betaGrid.min], "midpoint of bin at origin" );
   SET( "BETA_WID", betaGrid.stride, "uniform bin width" );
   SET( "PSI_REF", psiGrid[psiGrid.min], "midpoint of bin at origin" );
   SET( "PSI_WID", psiGrid.stride, "uniform bin width" );

   data = array->data();
   rowPtr = data->real32Data();
   scale = 0;

   struct CanonicalCrossPsf *crossPSF = 0;
   crossPSF = calServer.getAtom(crossPSF);
   CalReal64Vector xdspBins( 0, xdspGrid.length() );
   for ( int i = xdspGrid.min; i <= xdspGrid.max; ++i )
      xdspBins[i-xdspGrid.min] = xdspGrid[i];

   for ( int psi = psiGrid.min; psi <= psiGrid.max; ++psi )
      for ( int beta = betaGrid.min; beta <= betaGrid.max; ++beta )
	 copy( crossPSF->probability(0,betaGrid[beta],psiGrid[psi],
				     xdspBins,xdspGrid.stride),
	       size[2], rowPtr, scale );

   calServer.releaseAtom(crossPSF);

   SET( "SCALE", scale, "inverse normalizer" );
   standardize( data, 1/scale );
   delete data;

   // XDSP_RESPONSE_LA tensor (large angle scattering)

   size[0] = psiGrid.length();
   size[1] = wav1Grid.length();
   size[2] = xdspGrid.length();
   array = mcrset->addArray( "XDSP_RESPONSE_LA", Dal::Real32, size, "",
			     "response(xdsp,lambda,psi)" );

   SET( "XDSP_REF", xdspGrid[xdspGrid.min], "midpoint of bin at origin" );
   SET( "XDSP_WID", xdspGrid.stride, "uniform bin width" );
   SET( "WAV_REF", wav1Grid[wav1Grid.min], "midpoint of bin at origin" );
   SET( "WAV_WID", wav1Grid.stride, "uniform bin width" );
   SET( "PSI_REF", psiGrid[psiGrid.min], "midpoint of bin at origin" );
   SET( "PSI_WID", psiGrid.stride, "uniform bin width" );

   data = array->data();
   rowPtr = data->real32Data();
   scale = 0;

   for ( int psi = psiGrid.min; psi <= psiGrid.max; ++psi )
      for ( int wav = wav1Grid.min; wav <= wav1Grid.max; ++wav ) {
	 float energy = hc / wav1Grid[wav], dpsi = psiGrid[psi];

	 struct Spectrum dist;
	 dist.resize( xdspGrid.length() );
	 for ( int i = xdspGrid.min; i <= xdspGrid.max; ++i )
	    dist[i-xdspGrid.min]
	       = empiricalCrossLA(energy,dpsi,xdspGrid[i-.5],xdspGrid[i+.5]);

	 copy( dist, size[2], rowPtr, scale );
      }

   SET( "SCALE", scale, "inverse normalizer" );
   standardize( data, 1/scale );
   delete data;

   // PH_RESPONSE tensor

   size[0] = 18;
   size[1] = wav2Grid.length();
   size[2] = 4096;
   array = mcrset->addArray( "PH_RESPONSE", Dal::Real32, size, "",
			     "response(ph,lambda,node)" );

   SET( "WAV_REF", wav2Grid[wav2Grid.min], "midpoint of bin at origin" );
   SET( "WAV_WID", wav2Grid.stride, "uniform bin width" );

   data = array->data();
   rowPtr = data->real32Data();
   scale = 0;

   for ( unsigned ccd = 1; ccd <= 9; ++ccd ) {
      calServer.state()->ccd()->set( ccd );
      struct CanonicalRedist *redist = 0; redist = calServer.getAtom(redist);

      for ( unsigned node = 0; node < 2; ++node )
	 for ( int wav = wav2Grid.min; wav <= wav2Grid.max; ++wav ) {
	    float energy = hc/wav2Grid[wav];

	    float qe[10];
	    ccd_qe( qe, energy );
	    float ccdqe = qe[ccd];

	    struct Spectrum dist( redist->spectrum(energy*1000) );
	    Spectrum::iterator iter = dist.begin();
	    while ( iter != dist.end() ) *iter++ *= ccdqe;

	    copy( dist, size[2], rowPtr, scale );
	 }

      calServer.releaseAtom(redist);
   }

   SET( "SCALE", scale, "inverse normalizer" );
   standardize( data, 1/scale );
   delete data;

   // done

   dataSetServer->close( mcrset );

   calibFree();
}

#undef SET
